// PwdHash.cpp : Defines the entry point for the console application.
//

#include "stdafx.h"

#include <iostream>
#include <string>
#include <list>
#include "CBase64.h"
#include "md5.h"
#include "HashedPassword.h"
#include <cstringt.h>

using namespace std;


#define PASSWORD_PREFIX "@@"


char *HashedPassword::shorten(const char *source, int sourceLen, int shortenBy){ 
	int newLen = sourceLen - shortenBy;
	char *result =     new char[newLen];
	for(int i=0; i<newLen; i++){
		result[i] = source[i];
	}
	result[newLen-1] = '\0';

	return result; 
}

const char* HashedPassword::b64_hmac_md5(const char *key, const char* msg)
{
	char binary[16];
	hmac_md5((unsigned char*)msg, (int)strlen(msg), (unsigned char*)key, (int)strlen(key), binary);

	// Convert hash to base 64 
	char* str;
	CBase64 B64;
	int size = B64.CreateMatchingEncodingBuffer(sizeof(binary), &str);
	B64.EncodeBuffer(binary, sizeof(binary), str);


	cout << "Starting with: " << str << endl;; 

	if(str[size-2] == '='){
		cout << "Shorten by 2" << endl;
		char *tmp = str;
		str = shorten(str, size, 2);
		delete[] tmp;
	}else if(str[size-1] == '='){ 
		cout << "Shorten by 1" << endl;
		char *tmp = str;
		str = shorten(str, size, 1);
		delete[] tmp;
	}

	cout << "Returning: " << str << endl; 

	return str;
}

int HashedPassword::whitespaceIndex(const char *str){
	for(int i = 0; i<strlen(str); i++){
		if(!isalnum(str[i]) && str[i] != '_')
			return i;
	}

	return -1; 
}

bool HashedPassword::containsWhitespace(const char *str){
	return(whitespaceIndex(str) != -1);
}

string HashedPassword::nextExtra(string &extras){
	string result = "";


	if(extras.size() > 0){ 
		result += extras[0];
		extras = extras.substr(1);
	}

	return result;
}

void HashedPassword::rotate(string &arr, int amount){
	for(int i=0; i<amount; i++){
		int temp = arr[0]; 
		arr = arr.substr(1);
		arr += temp;
	}
}

int HashedPassword::between(int min, int interval, int offset){
	return min + offset % interval;
}

string HashedPassword::nextBetween(char base, int interval, string& extras){ 
	string extra = nextExtra(extras);

	int offset = (extra == "" ? 0 : extra[0]);
	char ch = (char) between(base, interval, offset);

	string result = "";
	result += ch; 

	return result;
}

bool HashedPassword::containsUpper(const char *str){
	for(int i=0; i<strlen(str); i++){
		if(isupper(str[i])) return true;
	}

	return false;
}

bool HashedPassword::containsLower(const char *str){ 
	for(int i=0; i<strlen(str); i++){
		if(islower(str[i])) return true;
	}

	return false;
}

bool HashedPassword::containsDigit(const char *str){
	for(int i=0; i<strlen(str); i++){
		if(isdigit(str[i])) return true; 
	}

	return false;
}

char *HashedPassword::applyConstraints(const char *hash, int size, bool nonalphanumeric){

	cout << "Applying constraints..." << endl;

	int startingSize = size - 4; 

	string result = "";
	for(int i=0; i<startingSize; i++){
		result += hash[i];
	}

	cout << "initial result: " << result << endl;

	string extras = ""; 
	if (startingSize >= 0)
	{
		for(int i = startingSize; i < strlen(hash); i++)
			extras += hash[i];
	}
	else
		extras = hash;
	string tmp = "";
	for (int i = 0; i < extras.length() - 1; i++)
	{
		tmp += extras.at(i);
		tmp += ",";
	}
	tmp += extras.at(extras.length() - 1);
	

	cout << "initial extras: " << extras << endl;

	result += (containsUpper(result.c_str ()) ? nextExtra(extras) : nextBetween('A', 26, extras));
	result += (containsLower(result.c_str()) ? nextExtra(extras) : nextBetween('a', 26, extras));
	cout << "after containsLower: " << result << endl; 

	result += (containsDigit(result.c_str()) ? nextExtra(extras) : nextBetween('0', 10, extras));
	cout << "after containsDigit: " << result << endl;

	result += ((containsWhitespace( result.c_str()) && nonalphanumeric) ? nextExtra(extras) : "+");

	cout << "Before while loop: " << result << endl;
	cout << "Extras before while loop: " << extras << endl; 

	while(containsWhitespace(result.c_str()) && !nonalphanumeric){
		cout << "Executing while loop!" << endl;
		int index = whitespaceIndex(result.c_str());
		result[index] = nextBetween('A', 26, extras)[0]; 
	}

	rotate(result, nextExtra(extras)[0]);

	char *final = new char[result.size() + 1];
	for(int i=0; i<result.size(); i++){
		final[i] = result[i];
	}
	final[result.size ()] = '\0';

	return final;
}

char *HashedPassword::getHashedPassword(const char *password, const char *realm){

	// Trim the password prefix off the front, if it's there
	if(!strncmp(PASSWORD_PREFIX, password, strlen(PASSWORD_PREFIX)))
		password += strlen(PASSWORD_PREFIX);

	const char *hash = b64_hmac_md5(password, realm);
	int size = strlen(password) + strlen(PASSWORD_PREFIX); 
	bool nonalphanumeric = containsWhitespace(password);
	char *result = applyConstraints(hash, size, nonalphanumeric);

	return result;
}

list<string> tokenize(string text, string sep)
{
	int last_pos=0;
	int pos=0;
	list<string> listWords;
	while(true)
	{
		pos=text.find_first_of(sep,last_pos);
		if(pos==string::npos)
		{
			if (last_pos < text.length())
				listWords.push_back(text.substr(last_pos,text.length()-last_pos));
			break;
		}
		else
		{
			string substr = text.substr(last_pos,pos-last_pos);
			if (substr.length() != 0)
				listWords.push_back(substr);
			last_pos=pos+1;
		}
	}
	return listWords;
}

string HashedPassword::ExtractDomain(string host)
{
	char* pMatch = "http://";
	int pos = host.find(pMatch);
	if (pos != -1)
		host = host.substr(pos + strlen(pMatch));

	pMatch = "https://";
	pos = host.find(pMatch);
	if (pos != -1)
		host = host.substr(pos + strlen(pMatch));

	pMatch = "/";
	pos = host.find(pMatch);
	host = host.substr(0, pos);

	string finalHost = "";
	list<string> strList = tokenize(host, ".");
	if (strList.size() > 2)
	{
		string str1 = strList.back();
		strList.pop_back();
		string str2 = strList.back();
		strList.pop_back();
		finalHost = str2 + "." + str1;
		string strDomains = "ab.ca|ac.ac|ac.at|ac.be|ac.cn|ac.il|ac.in|ac.jp|ac.kr|ac.nz|ac.th|ac.uk|ac.za|adm.br|adv.br|agro.pl|ah.cn|aid.pl|alt.za|am.br|arq.br|art.br|arts.ro|asn.au|asso.fr|asso.mc|atm.pl|auto.pl|bbs.tr|bc.ca|bio.br|biz.pl|bj.cn|br.com|cn.com|cng.br|cnt.br|co.ac|co.at|co.il|co.in|co.jp|co.kr|co.nz|co.th|co.uk|co.za|com.au|com.br|com.cn|com.ec|com.fr|com.hk|com.mm|com.mx|com.pl|com.ro|com.ru|com.sg|com.tr|com.tw|cq.cn|cri.nz|de.com|ecn.br|edu.au|edu.cn|edu.hk|edu.mm|edu.mx|edu.pl|edu.tr|edu.za|eng.br|ernet.in|esp.br|etc.br|eti.br|eu.com|eu.lv|fin.ec|firm.ro|fm.br|fot.br|fst.br|g12.br|gb.com|gb.net|gd.cn|gen.nz|gmina.pl|go.jp|go.kr|go.th|gob.mx|gov.br|gov.cn|gov.ec|gov.il|gov.in|gov.mm|gov.mx|gov.sg|gov.tr|gov.za|govt.nz|gs.cn|gsm.pl|gv.ac|gv.at|gx.cn|gz.cn|hb.cn|he.cn|hi.cn|hk.cn|hl.cn|hn.cn|hu.com|idv.tw|ind.br|inf.br|info.pl|info.ro|iwi.nz|jl.cn|jor.br|jpn.com|js.cn|k12.il|k12.tr|lel.br|ln.cn|ltd.uk|mail.pl|maori.nz|mb.ca|me.uk|med.br|med.ec|media.pl|mi.th|miasta.pl|mil.br|mil.ec|mil.nz|mil.pl|mil.tr|mil.za|mo.cn|muni.il|nb.ca|ne.jp|ne.kr|net.au|net.br|net.cn|net.ec|net.hk|net.il|net.in|net.mm|net.mx|net.nz|net.pl|net.ru|net.sg|net.th|net.tr|net.tw|net.za|nf.ca|ngo.za|nm.cn|nm.kr|no.com|nom.br|nom.pl|nom.ro|nom.za|ns.ca|nt.ca|nt.ro|ntr.br|nx.cn|odo.br|on.ca|or.ac|or.at|or.jp|or.kr|or.th|org.au|org.br|org.cn|org.ec|org.hk|org.il|org.mm|org.mx|org.nz|org.pl|org.ro|org.ru|org.sg|org.tr|org.tw|org.uk|org.za|pc.pl|pe.ca|plc.uk|ppg.br|presse.fr|priv.pl|pro.br|psc.br|psi.br|qc.ca|qc.com|qh.cn|re.kr|realestate.pl|rec.br|rec.ro|rel.pl|res.in|ru.com|sa.com|sc.cn|school.nz|school.za|se.com|se.net|sh.cn|shop.pl|sk.ca|sklep.pl|slg.br|sn.cn|sos.pl|store.ro|targi.pl|tj.cn|tm.fr|tm.mc|tm.pl|tm.ro|tm.za|tmp.br|tourism.pl|travel.pl|tur.br|turystyka.pl|tv.br|tw.cn|uk.co|uk.com|uk.net|us.com|uy.com|vet.br|web.za|web.com|www.ro|xj.cn|xz.cn|yk.ca|yn.cn|za.com";
		list<string> domainList = tokenize(strDomains, "|");
		while (domainList.size() != 0)
		{
			string domain = domainList.front();
			domainList.pop_front();
			if (finalHost.compare(domain) == 0)
			{
				finalHost = strList.back() + "." + domain;
				break;
			}
		}
	}
	else
	{
		finalHost = strList.front();
		strList.pop_front();
		while (strList.size() != 0)
		{
			string str = strList.front();
			strList.pop_front();
			finalHost += "." + str;
		}
	}

	return finalHost;
}

